/*
    *
    * Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
    *
    * This file is a part of DesQ project <https://gitlab.com/desq/>
    *

    *
    * This file was originally taken from LXQt project
    * <https://github.com/lxqt/lxqt-panel>
    * Several modifications have been made to suit the needs of DesQ
    * Any and all bug reports are to be filed with DesQ and not LXQt
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/

#include <QtCore>
#include <QDebug>
#include <QDBusConnectionInterface>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <cstdio>

#include "SNITypes.hpp"
#include "SNWatcher.hpp"

StatusNotifierWatcher::StatusNotifierWatcher( QObject *parent ) : QObject( parent ) {

    qRegisterMetaType<IconPixmap>( "IconPixmap" );
    qDBusRegisterMetaType<IconPixmap>();
    qRegisterMetaType<IconPixmapList>( "IconPixmapList" );
    qDBusRegisterMetaType<IconPixmapList>();
    qRegisterMetaType<ToolTip>( "ToolTip" );
    qDBusRegisterMetaType<ToolTip>();

    QDBusConnection dbus = QDBusConnection::sessionBus();
    serviceRunning = dbus.registerService( "org.kde.StatusNotifierWatcher" );
    objectRegistered = dbus.registerObject( "/StatusNotifierWatcher", this, QDBusConnection::ExportScriptableContents );

    mWatcher = new QDBusServiceWatcher( this );
    mWatcher->setConnection( dbus );
    mWatcher->setWatchMode( QDBusServiceWatcher::WatchForUnregistration );

    connect( mWatcher, &QDBusServiceWatcher::serviceUnregistered, this, &StatusNotifierWatcher::serviceUnregistered );
}

StatusNotifierWatcher::~StatusNotifierWatcher() {

    QDBusConnection::sessionBus().unregisterService( "org.kde.StatusNotifierWatcher" );
}

bool StatusNotifierWatcher::isServiceRunning() {

	return serviceRunning;
};

bool StatusNotifierWatcher::isObjectRegistered() {

	return objectRegistered;
};

void StatusNotifierWatcher::RegisterStatusNotifierItem( const QString &serviceOrPath ) {

    QString service = serviceOrPath;
    QString path = "/StatusNotifierItem";

    // workaround for sni-qt
    if ( service.startsWith( '/' ) ) {
        path = service;
        service = message().service();
    }

    QString notifierItemId = service + path;

    if ( QDBusConnection::sessionBus().interface()->isServiceRegistered( service ).value() && !mServices.contains( notifierItemId ) ) {
        mServices << notifierItemId;
        mWatcher->addWatchedService( service );
        emit StatusNotifierItemRegistered( notifierItemId );

        qDebug() << "New Status Icon" << notifierItemId;
    }
}

void StatusNotifierWatcher::RegisterStatusNotifierHost( const QString &service ) {

    if ( !mHosts.contains( service ) ) {
        mHosts.append( service );
        mWatcher->addWatchedService( service );

        qDebug() << "Host" << service << "registered";
    }
}

void StatusNotifierWatcher::serviceUnregistered( const QString &service ) {

    qDebug() << "Service" << service << "unregistered";

    mWatcher->removeWatchedService( service );

    if ( mHosts.contains( service ) ) {

        mHosts.removeAll( service );
        return;
    }

    QString match = service + '/';
    QStringList::Iterator it = mServices.begin();
    while ( it != mServices.end() ) {

        if ( it->startsWith( match ) ) {
            QString name = *it;
            it = mServices.erase( it );
            emit StatusNotifierItemUnregistered( name );
        }

        else
            ++it;
    }
}

int main( int argc, char *argv[] ) {

	QCoreApplication app( argc, argv );

    app.setOrganizationName( "DesQ" );
    app.setApplicationName( "DesQ Notfication Watcher" );
    app.setApplicationVersion( "1.0.0" );

    QCommandLineParser parser;
    parser.addHelpOption();         // Help
    parser.addVersionOption();      // Version

	parser.addOption( { { "d", "daemon" }, "Run this application as a daemon", "" } );
	parser.addOption( { "no-daemon", "Don't run this application as a daemon", "" } );

    parser.process( app );

    /*
        *
        * The reference used for writing this code into a daemon.
        * http://www.netzmafia.de/skripten/unix/linux-daemon-howto.html
        *
    */

    if ( not parser.isSet( "no-daemon" ) ) {

        pid_t pid = fork();

        if ( pid < 0 ) {

            // Something went wrong
            qDebug() << "Unable to create daemon process";

            // Exit the program
            exit( EXIT_FAILURE );
        }

        if ( pid > 0 ) {

            // Parent process. Inform the user and retire.
            qDebug() << "desq-snwatcher will continue running in the background.";

            // Exit the parent
            exit( EXIT_SUCCESS );
        }

        /* Change the file mode mask */
        umask( 0 );

        /* Create a new SID for the child process */
        pid_t sid = setsid();
        if ( sid < 0 ) {
            // TODO: Write to log file
            exit( EXIT_FAILURE );
        }

        /* Change to home folder */
        if ( chdir( QDir::homePath().toUtf8().data() ) < 0 ) {
            // TODO: Write to error log
            exit( EXIT_FAILURE );
        }

        /* Close the std file descriptors */
        close( STDIN_FILENO );
        close( STDOUT_FILENO );
        close( STDERR_FILENO );
    }

	/* Our actual code starts here */

	StatusNotifierWatcher *sniw = new StatusNotifierWatcher();

    /* Some one else has taken this service!! Grrr... */
	if ( not sniw->isServiceRunning() ) {
		qDebug() << "Unable to start the service 'org.kde.StatusNotifierWatcher'.";
        qDebug() << "Please ensure that another instance of this app is not running elsewhere.";
		return EXIT_FAILURE;
	}

    /* Something went wrong... */
	else if ( not sniw->isObjectRegistered() ) {
		qDebug() << "Unable to register the object '/StatusNotifierWatcher'.";
        qDebug() << "Please ensure that another instance of this app is not running elsewhere.";
		return EXIT_FAILURE;
	}

    /* All is well. */
	else {
		qDebug() << "Service org.kde.StatusNotifierWatcher running...";
		qDebug() << "Object /StatusNotifierWatcher registered...";
		qDebug() << "Ready for incoming connections...";
	}

    return app.exec();
};
